/*
	Sudoku generator

	Windows interface

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include "sudokures.h"

static HWND mainhwnd=NULL, printerhwnd=NULL, settingshwnd=NULL;
static HINSTANCE mainhInstance;

int nfilled=24;
int newnfilled, newsymmetrical;
int bestndigits=81;
int symmetrical=1;

#define THINWIDTH 1
#define THICKWIDTH 3
#define CELLSIZE 50

int displayboard[9][9];

int setboard[9][9];
int setboardsave[9][9];
int beforeboard[9][9];
int board[9][9];
int nmove=0;
int nsol=0;
int ntest=0;
int running=0;
int solved=0;

static HFONT font, oldfont;
static int default_textsize=CELLSIZE;
static int font_width, font_height;
static HANDLE printer_hDevMode, printer_hDevNames;
static RECT printer_lastMargin;
static int printer_marginunits;

static void setmenutext(int item, char *txt)
{
  MENUITEMINFO menuiteminfo;

  menuiteminfo.cbSize = sizeof(MENUITEMINFO);
  menuiteminfo.fMask = MIIM_STRING;
  menuiteminfo.fType = MFT_STRING;
  menuiteminfo.dwTypeData = txt;
  menuiteminfo.cch = strlen(txt);

  SetMenuItemInfo(GetMenu(mainhwnd), item, FALSE, &menuiteminfo);

  DrawMenuBar(mainhwnd);

}

setcreatemenutext(char *txt)
{
  setmenutext(IDM_MAIN_FILE_CREATE, txt);
}

setsolvemenutext(char *txt)
{
  setmenutext(IDM_MAIN_FILE_SOLVE, txt);
}

settitle(char *txt)
{
  SetWindowText(mainhwnd, txt);
}

static void generate_sudoku()
{
  bestndigits=81;
  running=1;
  ntest=0;
}

static printer_print(HWND hWnd)
{
  PRINTDLG pd;
  HDC hdc;
  HPEN hpen, hpenold;
  DOCINFO di;
  SIZE tsize;
  BOOL bSuccess = TRUE;
  BOOL bUserAbort = FALSE;
  int iPhysWidth, iPhysHeight, iPhysOffsetX, iPhysOffsetY;
  int mleft, mtop, mright, mbottom;
  int iLeftAdjust, iTopAdjust;
  int fontsize;
  int printer_font_width, printer_font_height;
  TEXTMETRIC tm;
  HFONT printer_font;
  HANDLE hold;
  int iColCopy, iNonColCopy, iPage, iLine, iCol, iChar;
  HBRUSH hbr, hbrold;
  LOGBRUSH logbrush;
  int oldmode;
  int printer_thinwidth, printer_thickwidth, printer_cellsize;
/*
  FILE *debug_fh;
*/
  int width, x, y, row, col, size;
  char txt[16];

  logbrush.lbStyle = BS_SOLID;
  logbrush.lbColor = RGB(0,0,0);
  logbrush.lbHatch = 0;

  memset(&pd, 0, sizeof(pd));
  pd.lStructSize = sizeof(pd);
  pd.hwndOwner = hWnd;
  pd.hDevMode = printer_hDevMode;
  pd.hDevNames = printer_hDevNames;
  pd.nMinPage = 1;
  pd.nMaxPage = 1;
  pd.nFromPage = 1;
  pd.nToPage = 1;
  pd.Flags = PD_RETURNDC;
  if(PrintDlg(&pd))
  {
    printer_hDevMode = pd.hDevMode;
    printer_hDevNames = pd.hDevNames;
    hdc = pd.hDC;
    mleft = MulDiv(printer_lastMargin.left, GetDeviceCaps(hdc, LOGPIXELSX), 1000);
    mtop = MulDiv(printer_lastMargin.top, GetDeviceCaps(hdc, LOGPIXELSY), 1000);
    mright = MulDiv(printer_lastMargin.right, GetDeviceCaps(hdc, LOGPIXELSX), 1000);
    mbottom = MulDiv(printer_lastMargin.bottom, GetDeviceCaps(hdc, LOGPIXELSY), 1000);
    iPhysOffsetX = GetDeviceCaps(hdc, PHYSICALOFFSETX);
    iPhysOffsetY = GetDeviceCaps(hdc, PHYSICALOFFSETY);
    iPhysWidth = GetDeviceCaps(hdc, PHYSICALWIDTH);
    iPhysHeight = GetDeviceCaps(hdc, PHYSICALHEIGHT);

    if(GetDeviceCaps(hdc, VERTRES)>GetDeviceCaps(hdc, HORZRES))
    {
      size = GetDeviceCaps(hdc, HORZRES);
    }
    else
    {
      size = GetDeviceCaps(hdc, VERTRES);
    }
    size = size*4/5;
    printer_cellsize = size/9;
    iTopAdjust = (GetDeviceCaps(hdc, VERTRES)-size)/2;
    iLeftAdjust = (GetDeviceCaps(hdc, HORZRES)-size)/2;
    fontsize = -size/10;
    printer_thinwidth = size/400;
    printer_thickwidth = printer_thinwidth*3;
/*
    debug_fh = fopen("sudoku.debug", "w");
    fprintf(debug_fh, "iPhysOffsetX: %d\n", iPhysOffsetX);
    fprintf(debug_fh, "iPhysOffsetY: %d\n", iPhysOffsetY);
    fprintf(debug_fh, "iPhysWidth: %d\n", iPhysWidth);
    fprintf(debug_fh, "iPhysHeight: %d\n", iPhysHeight);
    fprintf(debug_fh, "mleft: %d\n", mleft);
    fprintf(debug_fh, "mright: %d\n", mleft);
    fprintf(debug_fh, "mtop: %d\n", mleft);
    fprintf(debug_fh, "mbottom: %d\n", mleft);
    fprintf(debug_fh, "iLeftAdjust: %d\n", iLeftAdjust);
    fprintf(debug_fh, "iTopAdjust: %d\n", iTopAdjust);
    fprintf(debug_fh, "LOGPIXELSX: %d\n", GetDeviceCaps(hdc, LOGPIXELSX));
    fprintf(debug_fh, "LOGPIXELSY: %d\n", GetDeviceCaps(hdc, LOGPIXELSY));
    fprintf(debug_fh, "HORZRES: %d\n", GetDeviceCaps(hdc, HORZRES));
    fprintf(debug_fh, "VERTRES: %d\n", GetDeviceCaps(hdc, VERTRES));
    fprintf(debug_fh, "size: %d\n", size);
    fprintf(debug_fh, "fontsize: %d\n", fontsize);
    fclose(debug_fh);
*/
    SetMapMode(hdc, MM_TEXT);
    printer_font = CreateFont(fontsize,0,0,0,400,0,0,0,ANSI_CHARSET,OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,
		      PROOF_QUALITY,FIXED_PITCH|FF_MODERN,"Courier New");
    
    hold=SelectObject(hdc, printer_font);
    GetTextMetrics(hdc, &tm);
    GetTextExtentPoint32(hdc, "W", 1, &tsize);
    SelectObject(hdc, hold);
    printer_font_width = tsize.cx;
    printer_font_height = tsize.cy;

    memset(&di, 0, sizeof(di));
    di.cbSize = sizeof(di);
    di.lpszDocName = "Sudoku Print";
    oldmode=SetBkMode(hdc, TRANSPARENT);
    hbr = GetStockObject(NULL_BRUSH);
    hbrold = SelectObject(hdc, hbr);

    if(StartDoc(hdc, &di)>0)
    {
      for(iColCopy=0; iColCopy<((pd.Flags&PD_COLLATE)?pd.nCopies:1); iColCopy++)
      {
	iPage = 1;
	if((!(pd.Flags&PD_PAGENUMS)) ||
	   (iPage>=pd.nFromPage && iPage<=pd.nToPage))
	{
	  for(iNonColCopy = 0; iNonColCopy < ((pd.Flags&PD_COLLATE)?1:pd.nCopies); iNonColCopy++)
	  {
	    if(StartPage(hdc)<0)
	    {
	      bSuccess = FALSE;
	      break;
	    }
	    SetViewportOrgEx(hdc, iLeftAdjust, iTopAdjust, NULL);
	    y=0;
	    for(row=0;row<9;row++)
	    {
	      if(row>0)
	      {
		width = (row%3==0)?printer_thickwidth:printer_thinwidth;
		hpen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL,
				    width, &logbrush, 0, NULL);
		hpenold = SelectObject(hdc, hpen);
		MoveToEx(hdc, y, 0, NULL);
		LineTo(hdc, y, size);
		MoveToEx(hdc, 0, y, NULL);
		LineTo(hdc, size, y);
		SelectObject(hdc, hpenold);
		DeleteObject(hpen);
	      }
	      x=0;
	      for(col=0;col<9;col++)
	      {
		if(displayboard[row][col])
		{
		  sprintf(txt, "%d", abs(displayboard[row][col]));
		}
		else
		{
		  txt[0]=' ';
		  txt[1]=0;
		}
		oldfont=SelectObject(hdc, printer_font);
		SetTextColor(hdc, displayboard[row][col]<0?RGB(255,0,0):RGB(0,0,0));
		TextOut(hdc, x+(printer_cellsize-printer_font_width)/2, y, txt, 1);
		SelectObject(hdc, oldfont);
		x+=printer_cellsize;
	      }
	      y+=printer_cellsize;
	    }
	    hpen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL,
				printer_thickwidth, &logbrush, 0, NULL);
	    hpenold = SelectObject(hdc, hpen);
	    BeginPath(hdc);
	    MoveToEx(hdc, 0, 0, NULL);
	    LineTo(hdc, 0, size);
	    LineTo(hdc, size, size);
	    LineTo(hdc, size, 0);
	    CloseFigure(hdc);
	    EndPath(hdc);
	    StrokePath(hdc);
	    SelectObject(hdc, hpenold);
	    DeleteObject(hpen);
	    if(EndPage(hdc)<0)
	    {
	      bSuccess = FALSE;
	      break;
	    }
	  }
	  if(!bSuccess || bUserAbort) break;
	}
	if(!bSuccess || bUserAbort) break;
      }
      EndDoc(hdc);
    }

    SelectObject(hdc, hbrold);
    DeleteObject(hbr);
    SetBkMode(hdc, oldmode);
    DeleteObject(printer_font);
    DeleteDC(hdc);
  }
}

BOOL CALLBACK PrinterDlgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  int i;
  LPDEVNAMES lpdevnames;
  PAGESETUPDLG pdlg;
  switch(message)
  {
  case WM_INITDIALOG:
    lpdevnames = (LPDEVNAMES) GlobalLock(printer_hDevNames);
    SetDlgItemText(hWnd, IDM_PRINTER_NAME, ((char *) lpdevnames)+lpdevnames->wDeviceOffset);
    GlobalUnlock(lpdevnames);
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_PRINTER_SELECT:
      memset(&pdlg, 0, sizeof(pdlg));
      pdlg.lStructSize = sizeof(pdlg);
      pdlg.hwndOwner = mainhwnd;
      pdlg.hDevMode = printer_hDevMode;
      pdlg.hDevNames = printer_hDevNames;
      pdlg.rtMargin.top = printer_lastMargin.top;
      pdlg.rtMargin.bottom = printer_lastMargin.bottom;
      pdlg.rtMargin.left = printer_lastMargin.left;
      pdlg.rtMargin.right = printer_lastMargin.right;
      pdlg.Flags = PSD_MARGINS|printer_marginunits;
      PageSetupDlg(&pdlg);
      printer_hDevMode = pdlg.hDevMode;
      printer_hDevNames = pdlg.hDevNames;
      printer_lastMargin.top = pdlg.rtMargin.top;
      printer_lastMargin.bottom = pdlg.rtMargin.bottom;
      printer_lastMargin.left = pdlg.rtMargin.left;
      printer_lastMargin.right = pdlg.rtMargin.right;
      printer_marginunits = pdlg.Flags&(PSD_INTHOUSANDTHSOFINCHES|PSD_INHUNDREDTHSOFMILLIMETERS);
      lpdevnames = (LPDEVNAMES) GlobalLock(printer_hDevNames);
      SetDlgItemText(hWnd, IDM_PRINTER_NAME, ((char *) lpdevnames)+lpdevnames->wDeviceOffset);
      GlobalUnlock(lpdevnames);
      break;
    case IDM_PRINTER_PRINT:
      EndDialog(hWnd, IDM_PRINTER_PRINT);
      return TRUE;
    case IDM_PRINTER_CANCEL:
      EndDialog(hWnd, IDM_PRINTER_CANCEL);
      return TRUE;
    default:
      return TRUE;
    }
  default:
    return FALSE;
  }
  return FALSE;
}

BOOL CALLBACK SettingsDlgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  char tmp_label[80];
  int i;
  switch(message)
  {
  case WM_INITDIALOG:
    newnfilled = nfilled;
    newsymmetrical = symmetrical;
    SetDlgItemInt(hWnd, IDM_SETTINGS_FILLED, newnfilled, FALSE);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_SYMMETRICAL), BM_SETCHECK, newsymmetrical?BST_CHECKED:BST_UNCHECKED, 0);
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_SETTINGS_FILLED:
      newnfilled = GetDlgItemInt(hWnd, IDM_SETTINGS_FILLED, NULL, FALSE);
      break;
    case IDM_SETTINGS_SYMMETRICAL:
      newsymmetrical = SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_SYMMETRICAL), BM_GETCHECK, 0, 0)==BST_CHECKED;
      break;
    case IDM_SETTINGS_OK:
      EndDialog(hWnd, IDM_SETTINGS_OK);
      return TRUE;
    case IDM_SETTINGS_CANCEL:
      EndDialog(hWnd, IDM_SETTINGS_CANCEL);
      return TRUE;
    default:
      return TRUE;
    }
  default:
    return FALSE;
  }
  return FALSE;
}

static main_clear()
{
  int i,j;
  for(i=0;i<9;i++)
  for(j=0;j<9;j++)
  displayboard[i][j] = 0;
}

main_update()
{
  HDC hdc;
  HPEN hpen, hpenold;
  HBRUSH hbr, hbrold;
  char txt[16];
  int row,col,size,x,y,width;

  hdc = GetDC(mainhwnd);

  size = 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH-1;

  SetBkColor(hdc, RGB(255,255,255));

  y=0;
  for(row=0;row<9;row++)
  {
    if(row>0)
    {
      width = (row%3==0)?THICKWIDTH:THINWIDTH;
      hpen = CreatePen(PS_SOLID, width, RGB(0,0,0));
      hpenold = SelectObject(hdc, hpen);
      MoveToEx(hdc, y, 0, NULL);
      LineTo(hdc, y, size);
      MoveToEx(hdc, 0, y, NULL);
      LineTo(hdc, size, y);
      SelectObject(hdc, hpenold);
      DeleteObject(hpen);
      y+=width;
    }
    x=0;
    for(col=0;col<9;col++)
    {
      if(displayboard[row][col])
      {
        sprintf(txt, "%d", abs(displayboard[row][col]));
      }
      else
      {
        txt[0]=' ';
	txt[1]=0;
      }
      oldfont=SelectObject(hdc, font);
      SetTextColor(hdc, displayboard[row][col]<0?RGB(255,0,0):RGB(0,0,0));
      TextOut(hdc, x+(CELLSIZE-font_width)/2, y, txt, 1);
      SelectObject(hdc, oldfont);
      if(col>0) x+=(col%3==0)?THICKWIDTH:THINWIDTH;
      x+=CELLSIZE;
    }
    y+=CELLSIZE;
  }
  ReleaseDC(mainhwnd, hdc);
}

LRESULT CALLBACK mainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  HDC hdc;
  SIZE tsize;
  PAINTSTRUCT ps;
  OPENFILENAME openfile;
  char filename[200];
  char windowtitle[80];
  int status;

  local_srandom(time(NULL));

  switch(message)
  {
  case WM_PAINT:
    hdc = BeginPaint(hWnd, &ps);

    font = CreateFont(default_textsize,0,0,0,400,0,0,0,ANSI_CHARSET,OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,
		      DEFAULT_QUALITY,FIXED_PITCH|FF_MODERN,"Courier New");
    oldfont=SelectObject(hdc, font);
    GetTextExtentPoint32(hdc, "W", 1, &tsize);
    SelectObject(hdc, oldfont);
    font_width = tsize.cx;
    font_height = tsize.cy;
    EndPaint(hWnd, &ps);
    main_update();
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_MAIN_FILE_SETTINGS:
      if(DialogBox(mainhInstance, "SETTINGSDIALOG", mainhwnd, SettingsDlgProc)==IDM_SETTINGS_OK)
      {
        nfilled = newnfilled;
	symmetrical = newsymmetrical;
	main_clear();
        bestndigits=81;
	SetWindowText(mainhwnd, "Sudoku Generator");
	main_update();
      }
      break;
    case IDM_MAIN_FILE_PRINT:
      if(DialogBox(mainhInstance, "PRINTERDIALOG", mainhwnd, PrinterDlgProc)==IDM_PRINTER_PRINT)
      {
        printer_print(mainhwnd);
      }
      break;
    case IDM_MAIN_FILE_CREATE:
      if(running)
      {
	setmenutext(IDM_MAIN_FILE_CREATE, "Create");
	running=0;
	sprintf(windowtitle, "Sudoku Generator. Stopped at iteration %d, best: %d", ntest, bestndigits);
	SetWindowText(mainhwnd, windowtitle);
      }
      else
      {
	setmenutext(IDM_MAIN_FILE_CREATE, "Stop!");
	generate_sudoku();
	setmenutext(IDM_MAIN_FILE_SOLVE, "Solve");
      }
      break;
    case IDM_MAIN_FILE_SOLVE:
      if(!running)
      {
        if(solved)
	{
	  unsolve_sudoku();
	}
	else
	{
	  solve_sudoku();
	}
      }
      break;
    case IDM_MAIN_FILE_QUIT:
      exit(0);
      break;
    }
    break;
  case WM_DESTROY:
    DeleteObject(font);
    ExitProcess(0);
    break;
  default:
    return(DefWindowProc(hWnd, message, wParam, lParam));
  }
  return 0;
}

static printer_setup1st()
{
  PAGESETUPDLG pdlg;
  LPDEVNAMES lpdevnames;
  /* Set up default printer */
  memset(&pdlg, 0, sizeof(pdlg));
  pdlg.lStructSize = sizeof(pdlg);
  pdlg.hwndOwner = mainhwnd;
  pdlg.Flags = PSD_RETURNDEFAULT|PSD_MARGINS|PSD_INTHOUSANDTHSOFINCHES;
  pdlg.rtMargin.top = 500;
  pdlg.rtMargin.bottom = 500;
  pdlg.rtMargin.left = 500;
  pdlg.rtMargin.right = 500;
  PageSetupDlg(&pdlg);
  printer_lastMargin.top = pdlg.rtMargin.top;
  printer_lastMargin.bottom = pdlg.rtMargin.bottom;
  printer_lastMargin.left = pdlg.rtMargin.left;
  printer_lastMargin.right = pdlg.rtMargin.right;
  printer_marginunits = pdlg.Flags&(PSD_INTHOUSANDTHSOFINCHES|PSD_INHUNDREDTHSOFMILLIMETERS);
  printer_hDevMode = pdlg.hDevMode;
  printer_hDevNames = pdlg.hDevNames;
}


static void sudokumain()
{
  MSG Msg;
  BOOL status;

  printer_setup1st();

AGAIN:
  
  status=PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE);

AGAIN2:

  if(status)
  {
    TranslateMessage(&Msg);
    DispatchMessage(&Msg);
    goto AGAIN;
  }
  if(running)
  {
    sudoku_run();
    goto AGAIN;
  }
  status=GetMessage(&Msg, NULL, 0, 0);
  if(status == -1) exit(-1);
  if(status) goto AGAIN2;

  exit(0);
}

static BOOL mainInitApplication(HINSTANCE hInstance)
{
  WNDCLASS wc;
  BOOL rc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = mainWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = NULL;
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
  wc.lpszMenuName = "MAINMENU";
  wc.lpszClassName = "sudokumain";

  rc=RegisterClass(&wc);
  return rc;
}

static BOOL mainInitInstance(HINSTANCE hinstance, INT ncmdshow)
{
  RECT warea;

  main_clear();

  InitCommonControls();

  warea.left = 0;
  warea.right = 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH-1;
  warea.top = 0;
  warea.bottom = 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH-1;
  AdjustWindowRect(&warea, WS_OVERLAPPEDWINDOW, TRUE);

  mainhwnd = CreateWindow("sudokumain", "Sudoku Generator", WS_OVERLAPPEDWINDOW,
         25, 15, warea.right-warea.left+1, warea.bottom-warea.top, /* subtract one from height */
	 (HWND)NULL,(HMENU)NULL, hinstance, (LPSTR)NULL);

  if(mainhwnd == NULL) return 0;

  ShowWindow(mainhwnd, ncmdshow);
  UpdateWindow(mainhwnd);

  return 1;
}

int WINAPI WinMain(HINSTANCE hInst,HINSTANCE hPreInst, LPSTR lpszCmdLine,int nCmdShow)
{
  mainhInstance = hInst;

  mainInitApplication(mainhInstance);
  mainInitInstance(mainhInstance, nCmdShow);

  sudokumain();
}

